#-------------------------------------------------------------------------------
# Name:         BulkLoadPrepToolbox.pyt
# Purpose:      The bulk load prep toolbox contains custom Python tools that are
#               intended to increase the efficiency of data preparation prior to
#               bulk loading into Biotics
# Version:      1.2
# Author:       MMOORE, Pennsylvania Natural Heritage Program
# Created:      02/18/2018
# Updated:      01/02/2018
# Updated:      06/08/2020 by Whitney Weber (NatureServe) & Molly Moore to add REPLACE_SF & ADD_VISIT columns and fix typo in Aquatic analysis
#               09/16/2021 updated defaults to reflect new naming convention of point_projected, line_projected, poly_projected
#               09/28/2021 updated to add REPLACE_VISIT & V_ID columns since Dataset & Global ID fields are now recorded for Visits, hence we can determine when a Visit record should be replaced because it was edited
# Future Ideas:
#-------------------------------------------------------------------------------

#import libraries
import arcpy, os, datetime, time
from arcpy import env
from arcpy.sa import *

# set base workspace and printing variables
base_dir = os.path.dirname(__file__)
base_workspace = "in_memory"

# If running from Toolbox, gdb defined in base_workspace = os.path.join must exist, whereas when running it in DOS, the gdb will be automatically created.
# If you wish to run this file via command-line for slightly easier access and processing
# 1) scroll to the end of this file and uncomment the if __name__ == '__main__': and main() lines to make it run automatically
# 2) scroll to the main() function below and update inputs
# 2a) Pay attention to in_points (or lines/poly), eo_reps, and eo_sourceX parameters
# 2b) base_workspace = "in_memory" will result in faster performance until
#       memory limit is hit and then it will slow the entire thing to a crawl
#       Recommended you use a gdb for large datasets
# 2c) you can also update the debug flags below to print out timing information and full spam of row messages
# 3) open a command line window
# 4) cd <directory of this pyt file>  (base_dir is defined by the location of this script)
# 5) python BulkLoadPrepToolbox.pyt
# 5a) Use full path to python.exe if python not set globally
# 5b) ex: C:\Python27\ArcGIS10.6\python.exe BulkLoadPrepToolbox.pyt


# Debug flags for figuring out timing of things

# True if you want messages per row updated, but this will
#  slow the performance significantly potentially for large data
printRowMessages = False

# True if you want messages for timing values
printTimers = False

# True if you want messages for per-row timing values
printRowTimers = False

# Set value in seconds to only print cascade row timers that take longer than this
#  Not used if printRowTimers or printTimers are False
cascadeDelayThreshold = 8

# set environmental variables
arcpy.env.overwriteOutput = True
arcpy.env.qualifiedFieldNames = False
arcpy.env.workspace = base_workspace


# Adding main def for easy debugging
def main():
    tbx = Toolbox()
    tool = TerrestrialGrouping()

    # UPDATE BELOW THIS FOR COMMAND-LINE INPUTS
    #in_gdb = os.path.join(base_dir, 'ACCDC.gdb')
    in_gdb = os.path.join(r'C:\Temp\MJD\ACCDC', 'ACCDC.gdb')
    #in_gdb = os.path.join(base_dir, 'Testing_Small.gdb')
    #existing_gdb = os.path.join(base_dir, 'ACCDC.gdb')
    existing_gdb = os.path.join(r'C:\Temp\MJD\ACCDC', 'ACCDC.gdb')

    in_points = os.path.join(in_gdb, 'sf_point_locuncm_lt10000')
    #in_points = os.path.join(in_gdb, 'sf_point_locuncm_small')

    in_lines = None
    in_poly = None
    species_code = 'ELEMENT_SUBNATIONAL_ID'
    lu_separation = 'SEP_DISTANCE'
    loc_uncert = None
    loc_uncert_dist = 'LOCUNCM'
    eo_reps = os.path.join(existing_gdb, 'eo')
    eo_id_field = 'EO_ID'
    eo_sourcept = os.path.join(existing_gdb, 'source_pt')
    eo_sourceln = os.path.join(existing_gdb, 'source_line')
    eo_sourcepy = os.path.join(existing_gdb, 'source_poly')
    sf_id_field = 'SOURCE_FEA'
    eo_species_code_field = 'ELEMENT_SU'
    #base_workspace = os.path.join(base_dir, 'bulkLoadTempOutput.gdb')
    base_workspace = os.path.join(r'C:\Temp\MJD\ACCDC', 'bulkLoadTempOutput.gdb')
    runSourceFeatureAnalysis = False
    runGroupInputsToNewSourceFeaturesAnalysis = False
    printRowMessagesEveryX = 5000
    # UPDATE ABOVE THIS FOR COMMAND-LINE INPUTS

    arcpy.env.workspace = base_workspace
    params = tool.getParameterInfo()
    params[0].value = in_points
    params[1].value = in_lines
    params[2].value = in_poly
    params[3].value = species_code
    params[4].value = lu_separation
    params[5].value = loc_uncert
    params[6].value = loc_uncert_dist
    params[7].value = eo_reps
    params[8].value = eo_id_field
    params[9].value = eo_sourcept
    params[10].value = eo_sourceln
    params[11].value = eo_sourcepy
    params[12].value = sf_id_field
    params[13].value = eo_species_code_field
    params[14].value = base_workspace
    params[15].value = runSourceFeatureAnalysis
    params[16].value = runGroupInputsToNewSourceFeaturesAnalysis
    params[17].value = printRowMessagesEveryX

    tool.execute(params, None)
# end main

# Adding functions for starting timers and printing out end times
startTime = {}
def startTimer(id, skipPrint = False):
    if printTimers:
        startTime[id] = datetime.datetime.now()
        if skipPrint == False:
            arcpy.AddMessage("TIMER: Start Timer {} at {}".format(id, startTime[id]))

def endTimer(id, message, ifDelayGreater = None):
    if printTimers:
        endTime = datetime.datetime.now()
        duration = endTime - startTime[id]
        if id in startTime:
            del startTime[id]
        duration_s = duration.total_seconds()
        if ifDelayGreater is None:
            arcpy.AddMessage("TIMER: End Timer {} at {}: Took {} seconds for {}".format(id, endTime, duration_s, message))
        elif duration_s >= ifDelayGreater:
            arcpy.AddMessage("TIMER: End Timer {} at {}: Took {} seconds for {}. Delay greater than threshold of {} second".format(id, endTime, duration_s, message, ifDelayGreater))



class Toolbox(object):
    def __init__(self):
        self.label = "Biotics Bulk Load Toolbox"
        self.alias = "Biotics Bulk Load Toolbox"

        # List of tool classes associated with this toolbox
        self.tools = [TerrestrialGrouping,AquaticGrouping]

class TerrestrialGrouping(object):
    def __init__(self):
        self.label = "Terrestrial"
        self.description = """Used to prepare a feature class or shapefile for bulk load into Biotics by assigning an existing EOID/SFID or new EO/SF grouping string to observations based on separation distance."""
        self.canRunInBackground = False
        self.category = "EO/SF Separation Distance Analysis"

    def getParameterInfo(self):
        in_points = arcpy.Parameter(
            displayName = "Input POINT feature layer for separation distance analysis",
            name = "in_points",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        in_points.value = r"C:\Temp\Observations.gdb\point_projected"

        in_lines = arcpy.Parameter(
            displayName = "Input LINE feature layer for separation distance analysis",
            name = "in_lines",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input")
        in_lines.value = r"C:\Temp\Observations.gdb\line_projected"

        in_poly = arcpy.Parameter(
            displayName = "Input POLYGON feature layer for separation distance analysis",
            name = "in_poly",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input")
        in_poly.value = r"C:\Temp\Observations.gdb\poly_projected"

        species_code = arcpy.Parameter(
            displayName = "Species Identifier Field (values must match EO/SF species identifier field)",
            name = "species_code",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        species_code.value = r"est_id"
        species_code.parameterDependencies = [in_points.name]

        lu_separation = arcpy.Parameter(
            displayName = "Separation Distance Field",
            name = "lu_separation",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        lu_separation.value = r"SEP_DISTANCE"
        lu_separation.parameterDependencies = [in_points.name]

        loc_uncert = arcpy.Parameter(
            displayName = "Locational Uncertainty Type Field",
            name = "loc_uncert",
            datatype = "Field",
            parameterType = "Optional",
            direction = "Input")
        loc_uncert.value = r"loc_uncert"
        loc_uncert.parameterDependencies = [in_points.name]

        loc_uncert_dist = arcpy.Parameter(
            displayName = "Locational Uncertainty Distance Field (Units must be meters)",
            name = "loc_uncert_dist",
            datatype = "Field",
            parameterType = "Optional",
            direction = "Input")
        loc_uncert_dist.value = r"distance"
        loc_uncert_dist.parameterDependencies = [in_points.name]


        eo_reps = arcpy.Parameter(
            displayName = "Existing EO Reps Layer from Biotics backups or export",
            name = "eo_reps",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_reps.value = r"c:\Temp\BioticsBackups.gdb\eo"

        eo_id_field = arcpy.Parameter(
            displayName = "EO_ID Field used in EO Reps Layer",
            name = "eo_id_field",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        eo_id_field.value = "EO_ID"
        eo_id_field.parameterDependencies = [eo_reps.name]

        eo_sourcept = arcpy.Parameter(
            displayName = "Existing Source Point Layer from Biotics backups or export",
            name = "eo_sourcept",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_sourcept.value = r"c:\Temp\BioticsBackups.gdb\source_pt"

        eo_sourceln = arcpy.Parameter(
            displayName = "Existing Source Line Layer from Biotics backups or export",
            name = "eo_sourceln",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_sourceln.value = r"c:\Temp\BioticsBackups.gdb\source_line"

        eo_sourcepy = arcpy.Parameter(
            displayName = "Existing Source Polygon Layer from Biotics backups or export",
            name = "eo_sourcepy",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_sourcepy.value = r"c:\Temp\BioticsBackups.gdb\source_poly"

        sf_id_field = arcpy.Parameter(
            displayName = "SFID Field used in Source Points, Lines, and Polygon Layers",
            name = "sf_id_field",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        sf_id_field.value = r"SOURCE_FEA"
        sf_id_field.parameterDependencies = [eo_sourcept.name]

        eo_species_code_field = arcpy.Parameter(
            displayName = "Species Identifier Field (values must match input species identifier field)",
            name = "eo_species_code_field",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        eo_species_code_field.value = r"ELEMENT_SU"
        eo_species_code_field.parameterDependencies = [eo_reps.name]


        # Doing in_memory for very large datasets is problematic, so uncomment this and create an empty outputTest.gdb
        #  in the directory with this toolbox script
        base_workspace = arcpy.Parameter(
            displayName = "Temporary Output Workspace (in_memory faster, gdb more stable for large datasets)",
            name = "base_workspace",
            datatype = "DEWorkspace",
            #parameterType = "Optional",
            parameterType = "Required",
            direction = "Input")
        base_workspace.value = os.path.join(base_dir, 'bulkLoadTempOutput.gdb')

        # The only time SF_ID needs to be generated is if there are existing source features
        runSourceFeatureAnalysis = arcpy.Parameter(
            displayName = "Are there existing Source Features to consider?",
            name = "runSourceFeatureAnalysis",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input")
        runSourceFeatureAnalysis.value = True


        # The only time SF_NEW needs to be generated is if your data hasn't been pre-processed to group observations
        #  of same species near (within 9m) of each other into source features.
        #  By default, the bulk create tool will always create a single new source feature per feature in the input
        #  and so these steps are wasted processing for the majority of use cases.
        # NOTE: If this _is_ run, a manual process will need to be run to merge on that attribute before bulk
        #  create is ran.
        runGroupInputsToNewSourceFeaturesAnalysis = arcpy.Parameter(
            displayName = "Assign features within minimum mapping unit distance (9m) to same SF (SF_NEW)?",
            name = "runGroupInputsToNewSourceFeaturesAnalysis",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input")
        runGroupInputsToNewSourceFeaturesAnalysis.value = True

        # Log progress message every X rows, if set to 1 will print out every row,
        #   which was what it did before and is equivalent to setting printRowMessages = True
        printRowMessagesEveryX = arcpy.Parameter(
            displayName = "Log progress message every X rows (higher value improves performance)",
            name = "printRowMessagesEveryX",
            datatype = "GPLong",
            parameterType = "Optional",
            direction = "Input")
        printRowMessagesEveryX.value = 5000

        params = [in_points,in_lines,in_poly,species_code,lu_separation,loc_uncert,loc_uncert_dist,eo_reps,eo_id_field,eo_sourcept,eo_sourceln,eo_sourcepy,sf_id_field,eo_species_code_field, base_workspace, runSourceFeatureAnalysis, runGroupInputsToNewSourceFeaturesAnalysis, printRowMessagesEveryX]
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, params):
        return

    def updateMessages(self, params):
        return

    def execute(self, params, messages):
        startTimer("TerrestrialExecute")
        arcpy.AddMessage("""Welcome to the Source Feature and EO Assigner! This tool is designed to prepare a feature class or shapefile for bulk load into Biotics by assigning an existing or new SFID and EOID grouping variable to observations based on separation distance. This used to be done manually, so sit back and enjoy all the other work you can be doing instead of this!""")

        in_points = params[0].valueAsText
        in_lines = params[1].valueAsText
        in_poly = params[2].valueAsText
        in_species_code_field = params[3].valueAsText
        lu_separation_field = params[4].valueAsText
        loc_uncert_field = params[5].valueAsText
        loc_uncert_dist_field = params[6].valueAsText
        eo_reps = params[7].valueAsText
        eo_id_field = params[8].valueAsText
        eo_sourcept = params[9].valueAsText
        eo_sourceln = params[10].valueAsText
        eo_sourcepy = params[11].valueAsText
        sf_id_field = params[12].valueAsText
        eo_species_code_field = params[13].valueAsText

        # Doing in_memory for very large datasets is problematic, so uncomment this and create an empty outputTest.gdb
        #  in the directory with this toolbox script
        base_workspace = params[14].valueAsText

        # The only time source_feature IDs need to be generated is if there are existing source features
        #  or if you are using this to group source features with same species falling within 9m of each other.
        #  By default, the bulk create tool will always create a single new source feature per feature in the input
        #  and so these steps are wasted processing for the majority of use cases.
        runSourceFeatureAnalysis = params[15].value

        runGroupInputsToNewSourceFeaturesAnalysis = params[16].value

        printRowMessagesEveryX = params[17].value

        # if not in_memory, check if gdb and exists, then create it if not, and otherwise create folder
        if base_workspace != "in_memory":
            if base_workspace.endswith(".gdb") and not arcpy.Exists(base_workspace):
                splitPath = os.path.split(base_workspace)
                arcpy.AddMessage("Creating output workspace at {}".format(base_workspace))
                arcpy.CreateFileGDB_management(splitPath[0], splitPath[1])
            elif not os.path.exists(base_workspace):
                arcpy.AddMessage("Creating output workspace at {}".format(base_workspace))
                os.mkdir(base_workspace)


        arcpy.env.workspace = base_workspace

        arcpy.AddMessage("Preparing input data")
        #prepare single fc from biotics sf fcs
        startTimer("mergeBufferOriginal")
        sfs_in = [eo_sourcept,eo_sourceln,eo_sourcepy]
        sfs_out = ["eo_sourcept1","eo_sourceln1","eo_sourcepy1"]
        for sf_in,sf_out in zip(sfs_in,sfs_out):
            arcpy.Buffer_analysis(sf_in,sf_out,1)
        sf_merge = arcpy.Merge_management(sfs_out, "sf_merge")
        sf_lyr = arcpy.MakeFeatureLayer_management(sf_merge, "sf_lyr")
        # clean up the memory layers if used
        if base_workspace == "in_memory":
            for sf_o in sfs_out:
                arcpy.Delete_management(os.path.join(base_workspace, sf_o))
        endTimer("mergeBufferOriginal", "merging buffered original EO sources")

        data_in = []
        data_out = []
        if in_points:
            data_in.append(in_points)
            data_out.append("buff_pts")
        if in_lines:
            data_in.append(in_lines)
            data_out.append("buff_lines")
        if in_poly:
            data_in.append(in_poly)
            data_out.append("buff_polys")

        # Buffer inputs by LUC or 1 and merge
        startTimer("mergeBufferInput")
        join_id = 1
        for i,o in zip(data_in,data_out):
            startTimer("bufferInputAddJoinId")
            arcpy.AddField_management(i,"join_id","TEXT")
            with arcpy.da.UpdateCursor(i,"join_id") as cursor:
                for row in cursor:
                    row[0]=str(join_id)
                    cursor.updateRow(row)
                    join_id+=1
            arcpy.AddField_management(i,"buff_dist","SHORT")
            endTimer("bufferInputAddJoinId", "adding join ids")

            startTimer("bufferInput")
            if loc_uncert_dist_field:
                # if they give buffer dist with no type, need special case flag for later
                loc_uncert_none = False
                if loc_uncert_field is None:
                    loc_uncert_field = "OBJECTID"
                    loc_uncert_none = True

                with arcpy.da.UpdateCursor(i,["buff_dist",loc_uncert_field,loc_uncert_dist_field]) as cursor:
                    for row in cursor:
                        # if they give buffer dist with no type, assume all are estimated
                        loc_uncert_val = row[1]
                        if loc_uncert_none:
                            loc_uncert_val = "estimated"
                        if loc_uncert_val is None:
                            row[0] = 1
                            cursor.updateRow(row)
                        elif loc_uncert_val.lower() == "estimated":
                            if row[2] == 0:
                                row[2] = 1
                            row[0] = int(float(row[2])) + 1
                            cursor.updateRow(row)
                        else:
                            row[0] = 1
                            cursor.updateRow(row)
            else:
                with arcpy.da.UpdateCursor(i,"buff_dist") as cursor:
                    for row in cursor:
                        row[0] = 1
            ##arcpy.Buffer_analysis(i,o,"buff_dist")
            arcpy.CreateFeatureclass_management (base_workspace, o, "POLYGON", i, spatial_reference = i)
            flds = [f.name for f in arcpy.ListFields(i)] #use arcpy.ListFields (), manual entry, or some other means to determine shared fields
            flds += ["SHAPE@"] #add geom token to fields
            oid_field = arcpy.Describe(i).OIDFieldName

            #iterate input, get geometry, insert into output
            arcpy.AddMessage("Buffering input layer " + i)
            with arcpy.da.SearchCursor (i, flds) as sCurs:
                with arcpy.da.InsertCursor(o,flds) as iCurs:
                    count = 1
                    for row in sCurs:
                        if printRowMessages:
                            arcpy.AddMessage("Buffering OID "+str(row[flds.index(oid_field)]))
                        row = list(row)
                        geom = row [-1] #get geom object
                        buff_dist = flds.index("buff_dist")
                        b = geom.buffer(row[buff_dist])
                        row [-1] = b #add buffer to row
                        iCurs.insertRow (row)
                        if printRowMessagesEveryX > 0 and count % printRowMessagesEveryX == 0:
                            arcpy.AddMessage("Buffered to row "+str(count))
                        count += 1
                    arcpy.AddMessage("Finished buffering up to row "+str(count))

            del iCurs
            del sCurs
            endTimer("bufferInput", "buffering input layer {} into {}".format(i,o))
        # end for i,o in zip(data_in,data_out)

        data_merge = arcpy.Merge_management(data_out,"data_merge")
        # clean up the memory layers if used
        if base_workspace == "in_memory":
            for data_o in data_out:
                arcpy.Delete_management(os.path.join(base_workspace, data_o))

        # add index on in_species_code_field if on disk
        if base_workspace != "in_memory":
            startTimer("addInputSpeciesIndex")
            arcpy.AddIndex_management(data_merge, in_species_code_field, "SpeciesCodeIndex", "NON_UNIQUE", "ASCENDING")
            arcpy.AddSpatialIndex_management(data_merge)
            endTimer("addInputSpeciesIndex", "adding species index to merged input layer")

        # Make temporary feature layer out of the merge that'll be used for all selects and updates
        data_lyr = arcpy.MakeFeatureLayer_management(data_merge,"data_lyr")
        endTimer("mergeBufferInput", "merging buffered input layers into {}".format(data_lyr))

        #updated to account for double and float field types
        startTimer("addFields")
        if arcpy.ListFields(data_lyr,in_species_code_field)[0].type == 'Integer' or arcpy.ListFields(data_lyr,in_species_code_field)[0].type == 'Double' or arcpy.ListFields(data_lyr,in_species_code_field)[0].type == 'Float':
            species_query = "{}={}"
        else:
            species_query = "{}='{}'"
        if arcpy.ListFields(eo_reps,eo_species_code_field)[0].type == 'Integer' or arcpy.ListFields(eo_reps,eo_species_code_field)[0].type == 'Double' or arcpy.ListFields(eo_reps,eo_species_code_field)[0].type == 'Float':
            eo_species_query = "{}={}"
        else:
            eo_species_query = "{}='{}'"
        #get name of true OID field
        objectid_field = arcpy.Describe(data_lyr).OIDFieldName


        #add EO/SF ID fields if they do not already exist
        add_fields_text = ["SF_ID","SF_NEW","REPLACE_SF","V_ID","REPLACE_VISIT","ADD_VISIT","EO_ID","EO_NEW"]
        for field in add_fields_text:
            if len(arcpy.ListFields(data_lyr,field)) == 0:
                arcpy.AddField_management(data_lyr,field,"TEXT","","",500)
        add_fields_int = ["UNIQUEID"]
        for field in add_fields_int:
            if len(arcpy.ListFields(data_lyr,field)) == 0:
                arcpy.AddField_management(data_lyr,field,"LONG")
        endTimer("addFields", "adding fields to data_lyr")


        # add index on eo_species_code_field if not found and on disk
        if base_workspace != "in_memory":
            idIndexFound = False
            indexes = arcpy.ListIndexes(eo_reps)
            for index in indexes:
                if index.name == "EOSpeciesCodeIndex":
                    idIndexFound = True
                    break
            if idIndexFound == False:
                startTimer("addEoSpeciesIndex")
                arcpy.AddIndex_management(eo_reps, eo_species_code_field, "EOSpeciesCodeIndex", "NON_UNIQUE", "ASCENDING")
                endTimer("addEoSpeciesIndex", "adding species index to EO layer")


        #create feature layers to allow for selections
        eo_reps = arcpy.MakeFeatureLayer_management(eo_reps, "eo_reps")

        #set word index to assign words to new EO groups
        word_index = 1
        observation_num = 1

        arcpy.AddMessage("Beginning to assign EO IDs")
        startTimer("addEOIDs")
        #get total records in data_lyr for progress reporting messages
        total_obs = arcpy.GetCount_management(data_lyr)
        #start assigning loop
        search_fields = [objectid_field, "EO_ID", "EO_NEW", in_species_code_field, lu_separation_field]
        with arcpy.da.SearchCursor(data_lyr, search_fields) as cursor:
            for row in cursor:
                if printRowTimers:
                    startTimer("assignEOCascade", (cascadeDelayThreshold is not None and cascadeDelayThreshold != 0))
                objectid = row[0]
                #if one of the EOID fields already have a value, continue on to next feature
                if row[2] != None or (row[1] != None and row[1] != 0):
                    if printRowMessages:
                        arcpy.AddMessage("ObjectID " + str(objectid) + " EO Observation number " + str(observation_num) + "/" + str(total_obs) + " has already been assigned to a new or existing EO.")
                    pass
                else:
                    eo_species_code = row[3]
                    distance = str(row[4]*1000)+" METERS"
                    #arcpy.AddMessage("ObjectID " + str(objectid) + "; distance = " + distance)

                    #select feature and assign eo_species_code and separation distance variables
                    arcpy.SelectLayerByAttribute_management(data_lyr, "NEW_SELECTION", "{}={}".format(objectid_field,objectid))
                    #check for existing EO reps within separation distance of feature
                    arcpy.SelectLayerByAttribute_management(eo_reps, 'NEW_SELECTION', eo_species_query.format(eo_species_code_field,eo_species_code))
                    arcpy.SelectLayerByLocation_management(eo_reps, "WITHIN_A_DISTANCE", data_lyr, distance, "SUBSET_SELECTION")
                    #check for selection on eo_reps layer - if there is a selection, get eoid, select all observations within the separation distance, and assign existing eoid to selected features
                    selection_num = arcpy.Describe(eo_reps).fidset
                    if selection_num is not u'':
                        with arcpy.da.SearchCursor(eo_reps, eo_id_field) as cursor:
                            #eoid = sorted({row[0] for row in cursor}, reverse=True)[0] #use this if keeping newest EO
                            eoid = ",".join(sorted({str(int(row[0])) for row in cursor})) #use this if filling with EOIDs of all EOs within separation distance
                        #set arbitrary unequal counts to start while loop
                        countBefore = 0
                        countAfter = 1
                        while(countBefore!=countAfter):
                            countBefore = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                            arcpy.SelectLayerByLocation_management(data_lyr, "WITHIN_A_DISTANCE", data_lyr, distance, "ADD_TO_SELECTION")
                            arcpy.SelectLayerByAttribute_management(data_lyr, "SUBSET_SELECTION", species_query.format(in_species_code_field,eo_species_code))
                            countAfter = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                        # This only updates the ones selected
                        with arcpy.da.UpdateCursor(data_lyr, "EO_ID") as cursor:
                            for row in cursor:
                                row[0] = eoid
                                cursor.updateRow(row)
                        arcpy.SelectLayerByAttribute_management(data_lyr, "CLEAR_SELECTION")
                        if printRowMessages:
                            arcpy.AddMessage("ObjectID " + str(objectid)  + ", along with " + str(countAfter-1) + " observations were assigned an existing EO: " + str(eoid) + ". " + str(observation_num) + "/" + str(total_obs) + " completed.")
                    #if no existing EOs selected within separation distance, select all observations within the separation distance and assign new random word
                    else:
                        countBefore = 0
                        countAfter = 1
                        while(countBefore!=countAfter):
                            countBefore = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                            arcpy.SelectLayerByLocation_management(data_lyr, "WITHIN_A_DISTANCE", data_lyr, distance, "ADD_TO_SELECTION")
                            arcpy.SelectLayerByAttribute_management(data_lyr, "SUBSET_SELECTION", species_query.format(in_species_code_field,eo_species_code))
                            countAfter = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                        with arcpy.da.UpdateCursor(data_lyr, "EO_NEW") as cursor:
                            for row in cursor:
                                row[0] = "new_eo_" + str(word_index) #word_list[word_index]
                                cursor.updateRow(row)
                        if printRowMessages:
                            arcpy.AddMessage("ObjectID " + str(objectid)  + ", along with " + str(countAfter-1) + " observations were assigned a new EO: " + str(word_index) + ". " + str(observation_num) + "/" + str(total_obs) + " completed.")
                        word_index += 1
                observation_num += 1
                arcpy.SelectLayerByAttribute_management(data_lyr, "CLEAR_SELECTION")
                if printRowTimers:
                    endTimer("assignEOCascade", "assigning EO ID cascade for row " + str(observation_num), cascadeDelayThreshold)
                if printRowMessagesEveryX > 0 and observation_num % printRowMessagesEveryX == 0:
                    arcpy.AddMessage("Assigned EO IDs up to row "+str(observation_num))
            arcpy.AddMessage("Finished assigning EO IDs up to row "+str(observation_num))
        # end update all for EO_ID
        endTimer("addEOIDs", "updating data_lyr EO IDs")

        # The only time source_feature IDs need to be generated is if there are existing source features
        #  or if you are using this to group source features with same species falling within 9m of each other.
        #  By default, the bulk create tool will always create a single new source feature per feature in the input
        #  and so these steps are wasted processing for the majority of use cases.
        if runSourceFeatureAnalysis or runGroupInputsToNewSourceFeaturesAnalysis:
            arcpy.AddMessage("Beginning to assign SF IDs")
            startTimer("addSFIDs")
            observation_num = 1
            search_fields = [objectid_field, "SF_ID", "SF_NEW", in_species_code_field]
            with arcpy.da.SearchCursor(data_lyr, search_fields) as cursor:
                for row in cursor:
                    objectid = row[0]
                    if row[2] != None or (row[1] != None and row[1] != 0):
                        if printRowMessages:
                            arcpy.AddMessage("ObjectID " + str(objectid) + " SF Observation number " + str(observation_num) + "/" + str(total_obs) + " has already been assigned to a new or existing SF.")
                    else:
                        eo_species_code = row[3]

                        SF_ID_assigned = False
                        if runSourceFeatureAnalysis:
                            #check for existing SFs within 9m of feature (8m because of 1m buffer on SF layers)
                            if printRowTimers:
                                startTimer("assignExistingSFAssign", (cascadeDelayThreshold is not None and cascadeDelayThreshold != 0))
                            arcpy.SelectLayerByAttribute_management(data_lyr, "NEW_SELECTION", "{}={}".format(objectid_field, objectid))
                            #arcpy.AddMessage("eo_species_query: {}".format(eo_species_query.format(eo_species_code_field,eo_species_code)))
                            arcpy.SelectLayerByAttribute_management(sf_lyr, 'NEW_SELECTION', eo_species_query.format(eo_species_code_field,eo_species_code))
                            arcpy.SelectLayerByLocation_management(sf_lyr, "WITHIN_A_DISTANCE", data_lyr, "7 METERS", "SUBSET_SELECTION")
                            #check for selection on sf_merge layer - if there is a selection, get sfid, select all observations within the separation distance, and assign existing eoid to selected features
                            if arcpy.Describe('sf_lyr').fidset is not u'':
                                with arcpy.da.SearchCursor('sf_lyr', sf_id_field) as cursor:
                                    #sfid = sorted({row[0] for row in cursor}, reverse=True)[0] #use this line if you want to use the newest SF ID within separation distance
                                    sfid = ",".join(sorted({str(int(row[0])) for row in cursor})) # use this line if you want to list all SF IDs within separation distance
                                countBefore = 0
                                countAfter = 1
                                while(countBefore!=countAfter):
                                    countBefore = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                                    arcpy.SelectLayerByLocation_management(data_lyr, "WITHIN_A_DISTANCE", data_lyr, "7 METERS", "ADD_TO_SELECTION")
                                    arcpy.SelectLayerByAttribute_management(data_lyr, "SUBSET_SELECTION", species_query.format(in_species_code_field,eo_species_code))

                                    countAfter = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                                with arcpy.da.UpdateCursor(data_lyr, "SF_ID") as cursor:
                                    for row in cursor:
                                        row[0] = sfid
                                        cursor.updateRow(row)
                                if printRowMessages:
                                    arcpy.AddMessage("ObjectID " + str(objectid)  + ", along with " + str(countAfter-1) + " observations were assigned an existing SF: " + str(sfid) + ". " + str(observation_num) + "/" + str(total_obs) + " completed.")
                                SF_ID_assigned = True
                            if printRowTimers:
                                endTimer("assignExistingSFAssign", "assigning existing SF ID cascade for row " + str(observation_num), cascadeDelayThreshold)

                        if not SF_ID_assigned and runGroupInputsToNewSourceFeaturesAnalysis:
                            #if no existing SFs selected within separation distance, select all observations within the separation distance and assign new random word
                            if printRowTimers:
                                startTimer("assignSFCascade", (cascadeDelayThreshold is not None and cascadeDelayThreshold != 0))
                            countBefore = 0
                            countAfter = 1
                            while(countBefore!=countAfter):
                                countBefore = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                                arcpy.SelectLayerByLocation_management(data_lyr, "WITHIN_A_DISTANCE", data_lyr, "7 METERS", "ADD_TO_SELECTION")
                                arcpy.SelectLayerByAttribute_management(data_lyr, "SUBSET_SELECTION", species_query.format(in_species_code_field,eo_species_code))
                                countAfter = int(arcpy.GetCount_management("data_lyr").getOutput(0))
                            with arcpy.da.UpdateCursor(data_lyr, ["SF_NEW", "EO_NEW"]) as cursor:
                                for row in cursor:
                                    if row[1] != None:
                                        sf_id = "new_sf_"+str(word_index)
                                        row[0] = sf_id
                                    else:
                                        sf_id = "new_sf_"+str(word_index)
                                        row[0] = sf_id
                                    cursor.updateRow(row)
                            if printRowMessages:
                                arcpy.AddMessage("ObjectID " + str(objectid)  + ", along with " + str(countAfter-1) + " observations were assigned a new SF: " + sf_id + ". " + str(observation_num) + "/" + str(total_obs) + " completed.")
                            word_index += 1
                            if printRowTimers:
                                endTimer("assignSFCascade", "assigning new SF ID cascade for row " + str(observation_num), cascadeDelayThreshold)
                    observation_num += 1
                    arcpy.SelectLayerByAttribute_management(data_lyr, "CLEAR_SELECTION")
                    if printRowMessagesEveryX > 0 and observation_num % printRowMessagesEveryX == 0:
                        arcpy.AddMessage("Assigned SF IDs up to row "+str(observation_num))
                # end for SF ids loop
                arcpy.AddMessage("Finished assigning SF IDs up to row "+str(observation_num))
            endTimer("addSFIDs", "updating data_lyr SF_IDs")
            # end update all for SF_ID


            #create unique id value for each unique source feature
            startTimer("createUniqueId")
            i = 1
            with arcpy.da.SearchCursor(data_lyr, ["SF_ID", "SF_NEW", "UNIQUEID"]) as cursor:
                sfid1 = sorted({row[0] for row in cursor if row[0] is not None})
            with arcpy.da.SearchCursor(data_lyr, ["SF_ID", "SF_NEW", "UNIQUEID"]) as cursor:
                sfid2 = sorted({row[1] for row in cursor if row[0] is not None})
            sfid = sfid1 + sfid2
            sfid = [x for x in sfid if x is not None]
            sfid = [x.encode('UTF8') for x in sfid]
            for sf in sfid:
                with arcpy.da.UpdateCursor(data_lyr, ["SF_ID", "SF_NEW", "UNIQUEID"]) as cursor:
                    for row in cursor:
                        if row[0] == sf:
                            row[2] = i
                            cursor.updateRow(row)
                        elif row[1] == sf:
                            row[2] = i
                            cursor.updateRow(row)
                        else:
                            pass
                i += 1
            # end for unique sf ids update
            endTimer("createUniqueId", "assigning unique IDs to all source features")
        # end runSourceFeatureAnalysis

        startTimer("joinToInput")
        add_fields=add_fields_int+add_fields_text
        for data in data_in:
            arcpy.JoinField_management(data,"join_id",data_lyr,"join_id",add_fields)
            arcpy.DeleteField_management(data,"join_id")
            arcpy.DeleteField_management(data,"buff_dist")
        endTimer("joinToInput", "joining new fields to input layers")

        # clearing memory workspace from all generated during this
        if base_workspace == "in_memory":
            arcpy.Delete_management("in_memory")

        endTimer("TerrestrialExecute", "finishing terrestrial assign process")
        return
    # end def execute
# end class TerrestrialGrouping

class AquaticGrouping(object):
    def __init__(self):
        self.label = "Aquatic"
        self.description = """Used to prepare a feature class or shapefile representing aquatic species for bulk load into Biotics by assigning an existing EOID/SFID or new EO/SF grouping string to observations based on separation distance."""
        self.canRunInBackground = False
        self.category = "EO/SF Separation Distance Analysis"

    def getParameterInfo(self):
        in_points = arcpy.Parameter(
            displayName = "Input point feature layer for separation distance analysis",
            name = "in_points",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")

        in_lines = arcpy.Parameter(
            displayName = "Input line feature layer for separation distance analysis",
            name = "in_lines",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input")

        in_poly = arcpy.Parameter(
            displayName = "Input poly feature layer for separation distance analysis",
            name = "in_poly",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input")

        species_code = arcpy.Parameter(
            displayName = "Species identifier field used in input layers (values must match EO/SF species identifier field)",
            name = "species_code",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        species_code.parameterDependencies = [in_points.name]

        lu_separation = arcpy.Parameter(
            displayName = "Separation distance field (values must be in km)",
            name = "lu_separation",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        lu_separation.parameterDependencies = [in_points.name]

        eo_reps = arcpy.Parameter(
            displayName = "Existing EO reps layer",
            name = "eo_reps",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_reps.value = r"eo_reps"

        eo_id_field = arcpy.Parameter(
            displayName = "EOID field used in EO reps layer",
            name = "eo_id_field",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        eo_id_field.value = "EO_ID"
        eo_id_field.parameterDependencies = [eo_reps.name]

        eo_sourcept = arcpy.Parameter(
            displayName = "Existing source point layer",
            name = "eo_sourcept",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_sourcept.value = r"eosrc_pt"

        eo_sourceln = arcpy.Parameter(
            displayName = "Existing source line layer",
            name = "eo_sourceln",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_sourceln.value = r"eosrc_ln"

        eo_sourcepy = arcpy.Parameter(
            displayName = "Existing source polygon layer",
            name = "eo_sourcepy",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        eo_sourcepy.value = r"eosrc_py"

        sf_id_field = arcpy.Parameter(
            displayName = "SFID field used in source points, lines, and polygon layers",
            name = "sf_id_field",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        sf_id_field.value = "SF_ID"
        sf_id_field.parameterDependencies = [eo_sourcept.name]

        eo_species_code_field = arcpy.Parameter(
            displayName = "Species identifier field used in EO/SF layers (values must match input species identifier field)",
            name = "eo_species_code_field",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input")
        eo_species_code_field.parameterDependencies = [eo_reps.name]

        flowlines = arcpy.Parameter(
            displayName = "NHD flowlines",
            name = "flowlines",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input")
        flowlines.value = r'W:\Heritage\Heritage_Data\Heritage_Data_Tools\AquaticNetworkData.gdb\Aquatic_network\NHDFlowline'

        network = arcpy.Parameter(
            displayName = "Network dataset built on NHD flowlines",
            name = "network",
            datatype = "GPNetworkDatasetLayer",
            parameterType = "Required",
            direction = "Input")
        network.value = r'W:\Heritage\Heritage_Data\Heritage_Data_Tools\AquaticNetworkData.gdb\Aquatic_network\PA_network_ND'

        dams = arcpy.Parameter(
            displayName = "Dams/barrier points (must be snapped to NHD flowlines)",
            name = "dams",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input")

        snap_dist = arcpy.Parameter(
            displayName = "Snap distance in meters (distance to flowline beyond which observations will not be assigned/grouped)",
            name = "snap_dist",
            datatype = "GPString",
            parameterType = "Required",
            direction = "Input")
        snap_dist.value = "100"

        params = [in_points,in_lines,in_poly,species_code,lu_separation,eo_reps,eo_id_field,eo_sourcept,eo_sourceln,eo_sourcepy,sf_id_field,eo_species_code_field,flowlines,network,dams,snap_dist]
        return params

    def isLicensed(self):
        return True

    def updateParameters(self, params):
        return

    def updateMessages(self, params):
        return

    def execute(self, params, messages):
        arcpy.AddMessage("""Welcome to the Source Feature and EO Assigner! This tool is designed to prepare a feature class or shapefile for bulk load into Biotics by assigning an existing or new SFID and EOID grouping variable to observations based on separation distance. This used to be done manually, so sit back and enjoy all the other work you can be doing instead of this!""")

        in_points = params[0].valueAsText
        in_lines = params[1].valueAsText
        in_poly = params[2].valueAsText
        in_species_code_field = params[3].valueAsText
        lu_separation_field = params[4].valueAsText
        eo_reps = params[5].valueAsText
        eo_id_field = params[6].valueAsText
        eo_sourcept = params[7].valueAsText
        eo_sourceln = params[8].valueAsText
        eo_sourcepy = params[9].valueAsText
        sf_id_field = params[10].valueAsText
        eo_species_code_field = params[11].valueAsText
        flowlines = params[12].valueAsText
        network = params[13].valueAsText
        dams = params[14].valueAsText
        snap_dist = params[15].valueAsText

        arcpy.env.overwriteOutput = True
        arcpy.env.workspace = base_workspace

        arcpy.AddMessage("Preparing input data for use in EO/SF assignment.")
        data_in = []
        data_out = []
        if in_points:
            data_in.append(in_points)
            data_out.append("buff_pts")
        if in_lines:
            data_in.append(in_lines)
            data_out.append("buff_lines")
        if in_poly:
            data_in.append(in_poly)
            data_out.append("buff_polys")
        join_id1 = 1
        for i,o in zip(data_in,data_out):
            if len(arcpy.ListFields(i,"join_id")) == 0:
                arcpy.AddField_management(i,"join_id","TEXT")
            with arcpy.da.UpdateCursor(i,"join_id") as cursor:
                for row in cursor:
                    row[0]=str(join_id1)
                    cursor.updateRow(row)
                    join_id1+=1
            arcpy.FeatureVerticesToPoints_management(i,o,"ALL")
        species_pt = arcpy.Merge_management(data_out,"data_merge")
        species_pt_copy = arcpy.FeatureClassToFeatureClass_conversion(species_pt,base_workspace,"species_pt_copy")
        # clean up the memory layers if used
        if base_workspace == "in_memory":
            for data_o in data_out:
                arcpy.Delete_management(os.path.join(base_workspace, data_o))

        #prepare single fc from biotics sf fcs
        sfs_in = [eo_sourcept,eo_sourceln,eo_sourcepy]
        sfs_out = ["eo_sourcept1","eo_sourceln1","eo_sourcepy1"]
        for sf_in,sf_out in zip(sfs_in,sfs_out):
            arcpy.Buffer_analysis(sf_in,sf_out,1)
        sf_merge = arcpy.Merge_management(sfs_out, "sf_merge")
        # clean up the memory layers if used
        if base_workspace == "in_memory":
            for sf_o in sfs_out:
                arcpy.Delete_management(os.path.join(base_workspace, sf_o))

        # add index on eo_species_code_field if not found and on disk
        if base_workspace != "in_memory":
            idIndexFound = False
            indexes = arcpy.ListIndexes(sf_merge)
            for index in indexes:
                if index.name == "EOSpeciesCodeIndex":
                    idIndexFound = True
                    break
            if idIndexFound == False:
                startTimer("addEoSpeciesIndex")
                arcpy.AddIndex_management(sf_merge, eo_species_code_field, "EOSpeciesCodeIndex", "NON_UNIQUE", "ASCENDING")
                endTimer("addEoSpeciesIndex", "adding species index to EO layer")


        #delete identical points with tolerance to increase speed
        arcpy.DeleteIdentical_management(species_pt,["join_id","Shape"],"100 Meters")

        #add EO/SF ID fields if they do not already exist
        add_fields_text = ["SF_ID","SF_NEW","V_ID","REPLACE_SF","ADD_VISIT","EO_ID","EO_NEW"]
        for field in add_fields_text:
            if len(arcpy.ListFields(species_pt,field)) == 0:
                arcpy.AddField_management(species_pt,field,"TEXT","","",100)
        add_fields_int = ["UNIQUEID"]
        for field in add_fields_int:
            if len(arcpy.ListFields(species_pt,field)) == 0:
                arcpy.AddField_management(species_pt,field,"LONG")

        #create lookup dictionary of separation distances from lookup table
        lu_sep = {f[0]: f[1] for f in arcpy.da.SearchCursor(species_pt, [in_species_code_field,lu_separation_field])}
        #create list of species
        with arcpy.da.SearchCursor(species_pt,in_species_code_field) as cursor:
            species_list = sorted({row[0] for row in cursor})

        if arcpy.ListFields(species_pt,in_species_code_field)[0].type == 'Integer' or arcpy.ListFields(species_pt,in_species_code_field)[0].type == 'Double' or arcpy.ListFields(species_pt,in_species_code_field)[0].type == 'Float':
            species_query = "{}={}"
        else:
            species_query = "{}='{}'"
        if arcpy.ListFields(eo_reps,eo_species_code_field)[0].type == 'Integer' or arcpy.ListFields(eo_reps,eo_species_code_field)[0].type == 'Double' or arcpy.ListFields(eo_reps,eo_species_code_field)[0].type == 'Float':
            eo_species_query = "{}={}"
        else:
            eo_species_query = "{}='{}'"

        #separate buffered flowlines at dams
        if dams:
            #buffer dams by 1.1 meters
            dam_buff = arcpy.Buffer_analysis(dams,"dam_buff","1.1 Meter","FULL","FLAT")

        total_species = len(species_list)
        species_rep = 1
        group_id = 1
        for species in species_list:
            arcpy.AddMessage("Assigning EO for "+str(species_rep)+"/"+str(total_species)+": "+str(species))
            species_rep+=1
            s = arcpy.FeatureClassToFeatureClass_conversion(species_pt_copy,base_workspace,"s",species_query.format(in_species_code_field,species))
            eo = arcpy.MakeFeatureLayer_management(eo_reps,"eo",eo_species_query.format(eo_species_code_field,species))
            sf_lyr = arcpy.MakeFeatureLayer_management(sf_merge,"sf_lyr",eo_species_query.format(eo_species_code_field,species))
            for k,v in lu_sep.items():
                if k==species:
                    distance=(v*1000)-2

            #arcpy.AddMessage("Creating service area line layer for " +str(species) + " to compare to existing EOs")
            #create service area line layer
            eo_service_area_lyr = arcpy.na.MakeServiceAreaLayer(network,"eo_service_area_lyr","Length","TRAVEL_FROM",distance,polygon_type="NO_POLYS",line_type="TRUE_LINES",overlap="OVERLAP")
            eo_service_area_lyr = eo_service_area_lyr.getOutput(0)
            subLayerNames = arcpy.na.GetNAClassNames(eo_service_area_lyr)
            eo_facilitiesLayerName = subLayerNames["Facilities"]
            eo_serviceLayerName = subLayerNames["SALines"]
            arcpy.na.AddLocations(eo_service_area_lyr, eo_facilitiesLayerName, s, "", snap_dist)
            arcpy.na.Solve(eo_service_area_lyr)
            eo_lines = eo_service_area_lyr.listLayers(eo_serviceLayerName)[0]
            eo_flowline_clip = arcpy.CopyFeatures_management(eo_lines,"eo_service_area")
            eo_flowline_buff = arcpy.Buffer_analysis(eo_flowline_clip,"eo_flowline_buff","1 Meter","FULL","ROUND")
            eo_flowline_diss = arcpy.Dissolve_management(eo_flowline_buff,"eo_flowline_diss",multi_part="SINGLE_PART")
            eo_join = arcpy.SpatialJoin_analysis(target_features=eo_flowline_diss,join_features=eo,out_feature_class="eo_join",join_operation="JOIN_ONE_TO_MANY",join_type = "KEEP_ALL",match_option="INTERSECT")

            for k,v in lu_sep.items():
                if k==species:
                    distance=((v*1000/2)-2)

            #arcpy.AddMessage("Creating service area line layer for " +str(species) + " separation distance grouping")
            sp_service_area_lyr = arcpy.na.MakeServiceAreaLayer(network,"sp_service_area_lyr","Length","TRAVEL_FROM",
                distance,polygon_type="NO_POLYS",line_type="TRUE_LINES",overlap="OVERLAP")
            sp_service_area_lyr = sp_service_area_lyr.getOutput(0)
            subLayerNames = arcpy.na.GetNAClassNames(sp_service_area_lyr)
            sp_facilitiesLayerName = subLayerNames["Facilities"]
            sp_serviceLayerName = subLayerNames["SALines"]
            arcpy.na.AddLocations(sp_service_area_lyr, sp_facilitiesLayerName, s, "", snap_dist)
            arcpy.na.Solve(sp_service_area_lyr)
            sp_lines = sp_service_area_lyr.listLayers(sp_serviceLayerName)[0]
            sp_flowline_clip = arcpy.CopyFeatures_management(sp_lines,"sp_service_area")
            sp_flowline_buff = arcpy.Buffer_analysis(sp_flowline_clip,"sp_flowline_buff","1 Meter","FULL","ROUND")
            sp_flowline_diss = arcpy.Dissolve_management(sp_flowline_buff,"sp_flowline_diss",multi_part="SINGLE_PART")

            if dams:
                #split flowline buffers at dam buffers by erasing area of dam
                flowline_erase = arcpy.Erase_analysis(sp_flowline_diss,dam_buff,"flowline_erase")
                multipart_input = flowline_erase
                #multi-part to single part to create unique polygons
                single_part = arcpy.MultipartToSinglepart_management(multipart_input,"single_part")
            else:
                single_part = sp_flowline_diss

            #create unique group id
            arcpy.AddField_management(single_part,"group_id","LONG")
            with arcpy.da.UpdateCursor(single_part,"group_id") as cursor:
                for row in cursor:
                    row[0] = group_id
                    cursor.updateRow(row)
                    group_id+=1

            sp_join = arcpy.SpatialJoin_analysis(target_features=single_part,join_features=eo_join,out_feature_class="sp_join",join_operation="JOIN_ONE_TO_MANY",join_type="KEEP_ALL",match_option="INTERSECT")
            sp_join1 = arcpy.SpatialJoin_analysis(target_features=s,join_features=sp_join,out_feature_class="join",join_operation="JOIN_ONE_TO_MANY",join_type="KEEP_ALL",match_option="INTERSECT",search_radius="200 METERS")

            arcpy.AddField_management(sp_join1,"eoid","TEXT","","",100)
            #Create empty dictionaries
            Ndi = {}
            #create SearchCursor object
            with arcpy.da.SearchCursor(sp_join1, ["group_id",eo_id_field]) as cursor:
                for row in cursor:
                    if row[1]:
                        if not row[0] in Ndi:
                            Ndi[row[0]] = [row[1]]
                        else:
                            Ndi[row[0]].append(row[1])

            Ndi = {k: list(set(v)) for k, v in Ndi.items()}

            #create UpdateCursor
            if not Ndi == True:
                with arcpy.da.UpdateCursor (sp_join1, ["group_id","eoid"]) as cursor:
                    for row in cursor:
                        if row[0] in Ndi:
                            row[1] = ",".join(map(str, Ndi[row[0]]))
                            cursor.updateRow(row)

            arcpy.DeleteIdentical_management(sp_join1,["join_id","eoid","group_id"])

            id_fill = {f[0]: [f[1],f[2]] for f in arcpy.da.SearchCursor(sp_join1, ["join_id","group_id","eoid"])}

            with arcpy.da.UpdateCursor(species_pt,["join_id","EO_NEW","EO_ID"]) as cursor:
                for row in cursor:
                    for k,v in id_fill.items():
                        if k==row[0] and v[1] is not None:
                            row[2]=str(v[1])
                            cursor.updateRow(row)
                        elif k==row[0] and v[1] is None:
                            row[1]="new_eo_"+str(v[0])
                            cursor.updateRow(row)
                        else:
                            pass

        arcpy.DeleteIdentical_management(species_pt,["join_id","EO_ID","EO_NEW"])

        #get name of true OID field
        objectid_field = arcpy.Describe(species_pt).OIDFieldName

        # add index on in_species_code_field if on disk
        if base_workspace != "in_memory":
            startTimer("addInputSpeciesIndex")
            arcpy.AddIndex_management(species_pt, in_species_code_field, "SpeciesCodeIndex", "NON_UNIQUE", "ASCENDING")
            arcpy.AddSpatialIndex_management(species_pt)
            endTimer("addInputSpeciesIndex", "adding species index to merged input layer")

        species_lyr = arcpy.MakeFeatureLayer_management(species_pt,"species_lyr")

        arcpy.AddMessage("Assigning source feature IDs to all records.")
        search_fields = [objectid_field, "SF_ID", "SF_NEW", in_species_code_field]
        with arcpy.da.SearchCursor(species_lyr, search_fields) as cursor:
            for row in cursor:
                objectid = row[0]
                if row[2] != None and (row[1] != None or row[1] != 0):
                    pass
                else:
                    eo_species_code = row[3]

                    #check for existing SFs within 9m of feature (8m because of 1m buffer on SF layers)
                    arcpy.SelectLayerByAttribute_management(species_lyr, "NEW_SELECTION", "{}={}".format(objectid_field, objectid))
                    arcpy.SelectLayerByAttribute_management(sf_lyr, 'NEW_SELECTION', eo_species_query.format(eo_species_code_field,eo_species_code))
                    arcpy.SelectLayerByLocation_management(sf_lyr, "WITHIN_A_DISTANCE", species_lyr, "8 METERS", "SUBSET_SELECTION")
                    #check for selection on sf_merge layer - if there is a selection, get sfid, select all observations within the separation distance, and assign existing eoid to selected features
                    if arcpy.Describe('sf_lyr').fidset is not u'':
                        with arcpy.da.SearchCursor('sf_lyr', sf_id_field) as cursor:
                            #sfid = sorted({row[0] for row in cursor}, reverse=True)[0] #use this line if you want to use the newest SF ID within separation distance
                            sfid = ",".join(sorted({str(row[0]) for row in cursor})) # use this line if you want to list all SF IDs within separation distance
                        countBefore = 0
                        countAfter = 1
                        while(countBefore!=countAfter):
                            countBefore = int(arcpy.GetCount_management("species_lyr").getOutput(0))
                            arcpy.SelectLayerByLocation_management(species_lyr, "WITHIN_A_DISTANCE", species_lyr, "9 METERS", "ADD_TO_SELECTION")
                            arcpy.SelectLayerByAttribute_management(species_lyr, "SUBSET_SELECTION", species_query.format(in_species_code_field,eo_species_code))
                            countAfter = int(arcpy.GetCount_management("species_lyr").getOutput(0))
                        with arcpy.da.UpdateCursor(species_lyr, "SF_ID") as cursor:
                            for row in cursor:
                                row[0] = sfid
                                cursor.updateRow(row)
                    #if no existing SFs selected within separation distance, select all observations within the separation distance and assign new random word
                    else:
                        countBefore = 0
                        countAfter = 1
                        while(countBefore!=countAfter):
                            countBefore = int(arcpy.GetCount_management("species_lyr").getOutput(0))
                            arcpy.SelectLayerByLocation_management(species_lyr, "WITHIN_A_DISTANCE", species_lyr, "9 METERS", "ADD_TO_SELECTION")
                            arcpy.SelectLayerByAttribute_management(species_lyr, "SUBSET_SELECTION", species_query.format(in_species_code_field,eo_species_code))
                            countAfter = int(arcpy.GetCount_management("species_lyr").getOutput(0))
                        with arcpy.da.UpdateCursor(species_lyr, ["SF_NEW", "EO_NEW"]) as cursor:
                            for row in cursor:
                                if row[1] != None:
                                    sf_id = "new_sf_"+str(group_id)
                                    row[0] = sf_id
                                else:
                                    sf_id = "new_sf_"+str(group_id)
                                    row[0] = sf_id
                                cursor.updateRow(row)
                        group_id += 1
                arcpy.SelectLayerByAttribute_management(species_lyr, "CLEAR_SELECTION")

        #create unique id value for each unique source feature
        i = 1
        with arcpy.da.SearchCursor(species_lyr, ["SF_ID"]) as cursor:
            sfid1 = sorted({row[0] for row in cursor})
        with arcpy.da.SearchCursor(species_lyr, ["SF_NEW"]) as cursor:
            sfid2 = sorted({row[0] for row in cursor})
        sfid = sfid1 + sfid2
        sfid = [x for x in sfid if x is not None]
        for sf in sfid:
            with arcpy.da.UpdateCursor(species_lyr, ["SF_ID", "SF_NEW", "UNIQUEID"]) as cursor:
                for row in cursor:
                    if row[0] == sf:
                        row[2] = i
                        cursor.updateRow(row)
                    elif row[1] == sf:
                        row[2] = i
                        cursor.updateRow(row)
                    else:
                        pass
            i += 1

        add_fields=add_fields_int+add_fields_text
        for data in data_in:
            arcpy.JoinField_management(data,"join_id",species_lyr,"join_id",add_fields)

        for data in data_in:
            arcpy.DeleteField_management(data,"join_id")
        if base_workspace == "in_memory":
            arcpy.Delete_management("in_memory")

        return
# end class AquaticGrouping


# add a main declaration to allow easy debugging
# uncomment this to run from command-line
# IMPORTANT: If this is uncommented, trying to navigate to the folder containing it with ArcMap
#  will cause the arcCatalog browser to hang. Comment this out if you want to work with this folder
#  in ArcMap or else want to run it as an actual toolbox
# if __name__ == '__main__':
#    main()
